@@ -23,19 +23,35 @@
  * License for more details.
  */
 
-#define PY_SSIZE_T_CLEAN
-#include <Python.h>
-#include <string.h>
-
 #define PSYCOPG_MODULE
-#include "psycopg/config.h"
 #include "psycopg/psycopg.h"
+
 #include "psycopg/connection.h"
 #include "psycopg/cursor.h"
 #include "psycopg/pqpath.h"
 #include "psycopg/green.h"
 #include "psycopg/notify.h"
 
+#include <string.h>
+
+
+/* Return a new "string" from a char* from the database.
+ *
+ * On Py2 just get a string, on Py3 decode it in the connection codec.
+ *
+ * Use a fallback if the connection is NULL.
+ */
+PyObject *
+conn_text_from_chars(connectionObject *self, const char *str)
+{
+#if PY_MAJOR_VERSION < 3
+        return PyString_FromString(str);
+#else
+        const char *codec = self ? self->codec : "ascii";
+        return PyUnicode_Decode(str, strlen(str), codec, "replace");
+#endif
+}
+
 /* conn_notice_callback - process notices */
 
 static void
@@ -53,7 +69,15 @@ conn_notice_callback(void *args, const char *message)
     */
     notice = (struct connectionObject_notice *)
         malloc(sizeof(struct connectionObject_notice));
+    if (NULL == notice) {
+        /* Discard the notice in case of failed allocation. */
+        return;
+    }
     notice->message = strdup(message);
+    if (NULL == notice->message) {
+        free(notice);
+        return;
+    }
     notice->next = self->notice_pending;
     self->notice_pending = notice;
 }
@@ -77,8 +101,7 @@ conn_notice_process(connectionObject *self)
 
     while (notice != NULL) {
         PyObject *msg;
-        msg = PyString_FromString(notice->message);
-
+        msg = conn_text_from_chars(self, notice->message);
         Dprintf("conn_notice_process: %s", notice->message);
 
         /* Respect the order in which notices were produced,
@@ -146,8 +169,8 @@ conn_notifies_process(connectionObject *self)
                 (int) pgn->be_pid, pgn->relname);
 
         if (!(pid = PyInt_FromLong((long)pgn->be_pid))) { goto error; }
-        if (!(channel = PyString_FromString(pgn->relname))) { goto error; }
-        if (!(payload = PyString_FromString(pgn->extra))) { goto error; }
+        if (!(channel = conn_text_from_chars(self, pgn->relname))) { goto error; }
+        if (!(payload = conn_text_from_chars(self, pgn->extra))) { goto error; }
 
         if (!(notify = PyObject_CallFunctionObjArgs((PyObject *)&NotifyType,
                 pid, channel, payload, NULL))) {
@@ -213,38 +236,126 @@ conn_get_standard_conforming_strings(PGconn *pgconn)
     return equote;
 }
 
-/* Return a string containing the client_encoding setting.
+
+/* Remove irrelevant chars from encoding name and turn it uppercase.
  *
- * Return a new string allocated by malloc(): use free() to free it.
- * Return NULL in case of failure.
+ * Return a buffer allocated on Python heap,
+ * NULL and set an exception on error.
  */
 static char *
-conn_get_encoding(PGconn *pgconn)
+clean_encoding_name(const char *enc)
 {
-    const char *tmp, *i;
-    char *encoding, *j;
+    const char *i = enc;
+    char *rv, *j;
+
+    /* convert to upper case and remove '-' and '_' from string */
+    if (!(j = rv = PyMem_Malloc(strlen(enc) + 1))) {
+        PyErr_NoMemory();
+        return NULL;
+    }
+
+    while (*i) {
+        if (!isalnum(*i)) {
+            ++i;
+        }
+        else {
+            *j++ = toupper(*i++);
+        }
+    }
+    *j = '\0';
+
+    Dprintf("clean_encoding_name: %s -> %s", enc, rv);
+
+    return rv;
+}
+
+/* Convert a PostgreSQL encoding to a Python codec.
+ *
+ * Return a new copy of the codec name allocated on the Python heap,
+ * NULL with exception in case of error.
+ *
+ * 'enc' should be already normalized (uppercase, no - or _).
+ */
+static char *
+conn_encoding_to_codec(const char *enc)
+{
+    char *tmp;
+    Py_ssize_t size;
+    PyObject *pyenc = NULL;
+    char *rv = NULL;
+
+    /* Find the Py codec name from the PG encoding */
+    if (!(pyenc = PyDict_GetItemString(psycoEncodings, enc))) {
+        PyErr_Format(OperationalError,
+            "no Python codec for client encoding '%s'", enc);
+        goto exit;
+    }
+
+    /* Convert the codec in a bytes string to extract the c string. */
+    Py_INCREF(pyenc);
+    if (!(pyenc = psycopg_ensure_bytes(pyenc))) {
+        goto exit;
+    }
+
+    if (-1 == Bytes_AsStringAndSize(pyenc, &tmp, &size)) {
+        goto exit;
+    }
+
+    /* have our own copy of the python codec name */
+    rv = psycopg_strdup(tmp, size);
+
+exit:
+    Py_XDECREF(pyenc);
+    return rv;
+}
+
+/* Read the client encoding from the connection.
+ *
+ * Store the encoding in the pgconn->encoding field and the name of the
+ * matching python codec in codec. The buffers are allocated on the Python
+ * heap.
+ *
+ * Return 0 on success, else nonzero.
+ */
+static int
+conn_read_encoding(connectionObject *self, PGconn *pgconn)
+{
+    char *enc = NULL, *codec = NULL;
+    const char *tmp;
+    int rv = -1;
 
     tmp = PQparameterStatus(pgconn, "client_encoding");
     Dprintf("conn_connect: client encoding: %s", tmp ? tmp : "(none)");
     if (!tmp) {
         PyErr_SetString(OperationalError,
             "server didn't return client encoding");
-        return NULL;
+        goto exit;
     }
 
-    encoding = malloc(strlen(tmp)+1);
-    if (encoding == NULL) {
-        PyErr_NoMemory();
-        return NULL;
+    if (!(enc = clean_encoding_name(tmp))) {
+        goto exit;
     }
 
-    /* return in uppercase */
-    i = tmp;
-    j = encoding;
-    while (*i) { *j++ = toupper(*i++); }
-    *j = '\0';
+    /* Look for this encoding in Python codecs. */
+    if (!(codec = conn_encoding_to_codec(enc))) {
+        goto exit;
+    }
 
-    return encoding;
+    /* Good, success: store the encoding/codec in the connection. */
+    PyMem_Free(self->encoding);
+    self->encoding = enc;
+    enc = NULL;
+
+    PyMem_Free(self->codec);
+    self->codec = codec;
+    codec = NULL;
+
+    rv = 0;
+
+exit:
+    PyMem_Free(enc);
+    PyMem_Free(codec);
+    return rv;
 }
 
 int
@@ -299,6 +410,10 @@ conn_is_datestyle_ok(PGconn *pgconn)
     ds = PQparameterStatus(pgconn, "DateStyle");
     Dprintf("conn_connect: DateStyle %s", ds);
 
+    /* pgbouncer does not pass on DateStyle */
+    if (ds == NULL)
+      return 0;
+
     /* Return true if ds starts with "ISO"
      * e.g. "ISO, DMY" is fine, "German" not. */
     return (ds[0] == 'I' && ds[1] == 'S' && ds[2] == 'O');
@@ -320,9 +435,8 @@ conn_setup(connectionObject *self, PGconn *pgconn)
         PyErr_SetString(InterfaceError, "only protocol 3 supported");
         return -1;
     }
-    /* conn_get_encoding returns a malloc'd string */
-    self->encoding = conn_get_encoding(pgconn);
-    if (self->encoding == NULL) {
+
+    if (conn_read_encoding(self, pgconn)) {
         return -1;
     }
 
@@ -652,9 +766,7 @@ _conn_poll_setup_async(connectionObject *self)
             PyErr_SetString(InterfaceError, "only protocol 3 supported");
             break;
         }
-        /* conn_get_encoding returns a malloc'd string */
-        self->encoding = conn_get_encoding(self->pgconn);
-        if (self->encoding == NULL) {
+        if (conn_read_encoding(self, self->pgconn)) {
             break;
         }
         self->cancel = conn_get_cancel(self->pgconn);
@@ -748,7 +860,17 @@ conn_poll(connectionObject *self)
         if (res == PSYCO_POLL_OK && self->async_cursor) {
             /* An async query has just finished: parse the tuple in the
              * target cursor. */
-            cursorObject *curs = (cursorObject *)self->async_cursor;
+            cursorObject *curs;
+            PyObject *py_curs = PyWeakref_GetObject(self->async_cursor);
+            if (Py_None == py_curs) {
+                pq_clear_async(self);
+                PyErr_SetString(InterfaceError,
+                    "the asynchronous cursor has disappeared");
+                res = PSYCO_POLL_ERROR;
+                break;
+            }
+
+            curs = (cursorObject *)py_curs;
             IFCLEARPGRES(curs->pgres);
             curs->pgres = pq_get_last_result(self);
 
@@ -760,8 +882,7 @@ conn_poll(connectionObject *self)
             }
 
             /* We have finished with our async_cursor */
-            Py_XDECREF(self->async_cursor);
-            self->async_cursor = NULL;
+            Py_CLEAR(self->async_cursor);
         }
         break;
 
@@ -873,33 +994,54 @@ conn_set_client_encoding(connectionObject *self, const char *enc)
     PGresult *pgres = NULL;
     char *error = NULL;
     char query[48];
-    int res = 0;
+    int res = 1;
+    char *codec = NULL;
+    char *clean_enc = NULL;
 
     /* If the current encoding is equal to the requested one we don't
        issue any query to the backend */
     if (strcmp(self->encoding, enc) == 0) return 0;
 
+    /* We must know what python codec this encoding is. */
+    if (!(clean_enc = clean_encoding_name(enc))) { goto exit; }
+    if (!(codec = conn_encoding_to_codec(clean_enc))) { goto exit; }
+
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
 
     /* set encoding, no encoding string is longer than 24 bytes */
-    PyOS_snprintf(query, 47, "SET client_encoding = '%s'", enc);
+    PyOS_snprintf(query, 47, "SET client_encoding = '%s'", clean_enc);
 
     /* abort the current transaction, to set the encoding ouside of
        transactions */
-    res = pq_abort_locked(self, &pgres, &error, &_save);
-
-    if (res == 0) {
-        res = pq_execute_command_locked(self, query, &pgres, &error, &_save);
-        if (res == 0) {
-            /* no error, we can proceeed and store the new encoding */
-            if (self->encoding) free(self->encoding);
-            self->encoding = strdup(enc);
-            Dprintf("conn_set_client_encoding: set encoding to %s",
-                    self->encoding);
-        }
+    if ((res = pq_abort_locked(self, &pgres, &error, &_save))) {
+        goto endlock;
     }
 
+    if ((res = pq_execute_command_locked(self, query, &pgres, &error, &_save))) {
+        goto endlock;
+    }
+
+    /* no error, we can proceeed and store the new encoding */
+    {
+        char *tmp = self->encoding;
+        self->encoding = clean_enc;
+        PyMem_Free(tmp);
+        clean_enc = NULL;
+    }
+
+    /* Store the python codec too. */
+    {
+        char *tmp = self->codec;
+        self->codec = codec;
+        PyMem_Free(tmp);
+        codec = NULL;
+    }
+
+    Dprintf("conn_set_client_encoding: set encoding to %s (codec: %s)",
+            self->encoding, self->codec);
+
+endlock:
 
     pthread_mutex_unlock(&self->lock);
     Py_END_ALLOW_THREADS;
@@ -907,6 +1049,10 @@ conn_set_client_encoding(connectionObject *self, const char *enc)
     if (res < 0)
         pq_complete_error(self, &pgres, &error);
 
+exit:
+    PyMem_Free(clean_enc);
+    PyMem_Free(codec);
+
     return res;
 }
 
@@ -965,8 +1111,8 @@ conn_tpc_command(connectionObject *self, const char *cmd, XidObject *xid)
     Dprintf("conn_tpc_command: %s", cmd);
 
     /* convert the xid into PostgreSQL transaction id while keeping the GIL */
-    if (!(tid = xid_get_tid(xid))) { goto exit; }
-    if (!(ctid = PyString_AsString(tid))) { goto exit; }
+    if (!(tid = psycopg_ensure_bytes(xid_get_tid(xid)))) { goto exit; }
+    if (!(ctid = Bytes_AsString(tid))) { goto exit; }
 
     Py_BEGIN_ALLOW_THREADS;
     pthread_mutex_lock(&self->lock);
