@@ -23,21 +23,19 @@
  * License for more details.
  */
 
-#define PY_SSIZE_T_CLEAN
-#include <Python.h>
-#include <structmember.h>
-#include <stringobject.h>
+#define PSYCOPG_MODULE
+#include "psycopg/psycopg.h"
+
+#include "psycopg/connection.h"
+#include "psycopg/cursor.h"
+#include "psycopg/pqpath.h"
+#include "psycopg/lobject.h"
+#include "psycopg/green.h"
+#include "psycopg/xid.h"
 
 #include <string.h>
 #include <ctype.h>
 
-#define PSYCOPG_MODULE
-#include "psycopg/config.h"
-#include "psycopg/python.h"
-#include "psycopg/psycopg.h"
-#include "psycopg/connection.h"
-#include "psycopg/cursor.h"
-#include "psycopg/lobject.h"
 
 /** DBAPI methods **/
 
@@ -66,6 +64,21 @@ psyco_conn_cursor(connectionObject *self, PyObject *args, PyObject *keywds)
 
     EXC_IF_CONN_CLOSED(self);
 
+    if (self->status != CONN_STATUS_READY &&
+        self->status != CONN_STATUS_BEGIN &&
+        self->status != CONN_STATUS_PREPARED) {
+        PyErr_SetString(OperationalError,
+                        "asynchronous connection attempt underway");
+        return NULL;
+    }
+
+    if (name != NULL && self->async == 1) {
+        PyErr_SetString(ProgrammingError,
+                        "asynchronous connections "
+                        "cannot produce named cursors");
+        return NULL;
+    }
+
     Dprintf("psyco_conn_cursor: new cursor for connection at %p", self);
     Dprintf("psyco_conn_cursor:     parameters: name = %s", name);
 
@@ -73,7 +86,7 @@ psyco_conn_cursor(connectionObject *self, PyObject *args, PyObject *keywds)
     if (name)
         obj = PyObject_CallFunction(factory, "Os", self, name);
     else
-        obj = PyObject_CallFunction(factory, "O", self);
+        obj = PyObject_CallFunctionObjArgs(factory, self, NULL);
 
     if (obj == NULL) return NULL;
     if (PyObject_IsInstance(obj, (PyObject *)&cursorType) == 0) {
@@ -85,7 +98,7 @@ psyco_conn_cursor(connectionObject *self, PyObject *args, PyObject *keywds)
 
     Dprintf("psyco_conn_cursor: new cursor at %p: refcnt = "
         FORMAT_CODE_PY_SSIZE_T,
-        obj, obj->ob_refcnt
+        obj, Py_REFCNT(obj)
       );
     return obj;
 }
@@ -100,8 +113,6 @@ psyco_conn_close(connectionObject *self, PyObject *args)
 {
     EXC_IF_CONN_CLOSED(self);
 
-    if (!PyArg_ParseTuple(args, "")) return NULL;
-
     Dprintf("psyco_conn_close: closing connection at %p", self);
     conn_close(self);
     Dprintf("psyco_conn_close: connection at %p closed", self);
@@ -119,8 +130,8 @@ static PyObject *
 psyco_conn_commit(connectionObject *self, PyObject *args)
 {
     EXC_IF_CONN_CLOSED(self);
-
-    if (!PyArg_ParseTuple(args, "")) return NULL;
+    EXC_IF_CONN_ASYNC(self, commit);
+    EXC_IF_TPC_BEGIN(self, commit);
 
     if (conn_commit(self) < 0)
         return NULL;
@@ -139,8 +150,8 @@ static PyObject *
 psyco_conn_rollback(connectionObject *self, PyObject *args)
 {
     EXC_IF_CONN_CLOSED(self);
-
-    if (!PyArg_ParseTuple(args, "")) return NULL;
+    EXC_IF_CONN_ASYNC(self, rollback);
+    EXC_IF_TPC_BEGIN(self, rollback);
 
     if (conn_rollback(self) < 0)
         return NULL;
@@ -150,6 +161,227 @@ psyco_conn_rollback(connectionObject *self, PyObject *args)
 }
 
 
+#define psyco_conn_xid_doc \
+"xid(format_id, gtrid, bqual) -- create a transaction identifier."
+
+static PyObject *
+psyco_conn_xid(connectionObject *self, PyObject *args, PyObject *kwargs)
+{
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_TPC_NOT_SUPPORTED(self);
+
+    return PyObject_Call((PyObject *)&XidType, args, kwargs);
+}
+
+
+#define psyco_conn_tpc_begin_doc \
+"tpc_begin(xid) -- begin a TPC transaction with given transaction ID xid."
+
+static PyObject *
+psyco_conn_tpc_begin(connectionObject *self, PyObject *args)
+{
+    PyObject *rv = NULL;
+    XidObject *xid = NULL;
+    PyObject *oxid;
+
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, tpc_begin);
+    EXC_IF_TPC_NOT_SUPPORTED(self);
+
+    if (!PyArg_ParseTuple(args, "O", &oxid)) {
+        goto exit;
+    }
+
+    if (NULL == (xid = xid_ensure(oxid))) {
+        goto exit;
+    }
+
+    /* check we are not in a transaction */
+    if (self->status != CONN_STATUS_READY) {
+        PyErr_SetString(ProgrammingError,
+            "tpc_begin must be called outside a transaction");
+        goto exit;
+    }
+
+    /* two phase commit and autocommit make no point */
+    if (self->isolation_level == ISOLATION_LEVEL_AUTOCOMMIT) {
+        PyErr_SetString(ProgrammingError,
+            "tpc_begin can't be called in autocommit mode");
+        goto exit;
+    }
+
+    if (conn_tpc_begin(self, xid) < 0) {
+        goto exit;
+    }
+
+    Py_INCREF(Py_None);
+    rv = Py_None;
+
+exit:
+    Py_XDECREF(xid);
+    return rv;
+}
+
+
+#define psyco_conn_tpc_prepare_doc \
+"tpc_prepare() -- perform the first phase of a two-phase transaction."
+
+static PyObject *
+psyco_conn_tpc_prepare(connectionObject *self, PyObject *args)
+{
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, tpc_prepare);
+    EXC_IF_TPC_PREPARED(self, tpc_prepare);
+
+    if (NULL == self->tpc_xid) {
+        PyErr_SetString(ProgrammingError,
+            "prepare must be called inside a two-phase transaction");
+        return NULL;
+    }
+
+    if (0 > conn_tpc_command(self, "PREPARE TRANSACTION", self->tpc_xid)) {
+        return NULL;
+    }
+
+    /* transaction prepared: set the state so that no operation
+     * can be performed until commit. */
+    self->status = CONN_STATUS_PREPARED;
+
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+
+/* the type of conn_commit/conn_rollback */
+typedef int (*_finish_f)(connectionObject *self);
+
+/* Implement tpc_commit/tpc_rollback.
+ *
+ * This is a common framework performing the chechs and state manipulation
+ * common to the two functions.
+ *
+ * Parameters are:
+ * - self, args: passed by Python
+ * - opc_f: the function to call in case of one-phase commit/rollback
+ *          one of conn_commit/conn_rollback
+ * - tpc_cmd: the command to execute for a two-phase commit/rollback
+ *
+ * The function can be called in three cases:
+ * - If xid is specified, the status must be "ready";
+ *   issue the commit/rollback prepared.
+ * - if xid is not specified and status is "begin" with a xid,
+ *   issue a normal commit/rollback.
+ * - if xid is not specified and status is "prepared",
+ *   issue the commit/rollback prepared.
+ */
+static PyObject *
+_psyco_conn_tpc_finish(connectionObject *self, PyObject *args,
+    _finish_f opc_f, char *tpc_cmd)
+{
+    PyObject *oxid = NULL;
+    XidObject *xid = NULL;
+    PyObject *rv = NULL;
+
+    if (!PyArg_ParseTuple(args, "|O", &oxid)) { goto exit; }
+
+    if (oxid) {
+        if (!(xid = xid_ensure(oxid))) { goto exit; }
+    }
+
+    if (xid) {
+        /* committing/aborting a recovered transaction. */
+        if (self->status != CONN_STATUS_READY) {
+            PyErr_SetString(ProgrammingError,
+                "tpc_commit/tpc_rollback with a xid "
+                "must be called outside a transaction");
+            goto exit;
+        }
+        if (0 > conn_tpc_command(self, tpc_cmd, xid)) {
+            goto exit;
+        }
+    } else {
+        /* committing/aborting our own transaction. */
+        if (!self->tpc_xid) {
+            PyErr_SetString(ProgrammingError,
+                "tpc_commit/tpc_rollback with no parameter "
+                "must be called in a two-phase transaction");
+            goto exit;
+        }
+
+        switch (self->status) {
+          case CONN_STATUS_BEGIN:
+            if (0 > opc_f(self)) { goto exit; }
+            break;
+
+          case CONN_STATUS_PREPARED:
+            if (0 > conn_tpc_command(self, tpc_cmd, self->tpc_xid)) {
+                goto exit;
+            }
+            break;
+
+          default:
+            PyErr_SetString(InterfaceError,
+                "unexpected state in tpc_commit/tpc_rollback");
+            goto exit;
+        }
+
+        Py_CLEAR(self->tpc_xid);
+
+        /* connection goes ready */
+        self->status = CONN_STATUS_READY;
+    }
+
+    Py_INCREF(Py_None);
+    rv = Py_None;
+
+exit:
+    Py_XDECREF(xid);
+    return rv;
+}
+
+#define psyco_conn_tpc_commit_doc \
+"tpc_commit([xid]) -- commit a transaction previously prepared."
+
+static PyObject *
+psyco_conn_tpc_commit(connectionObject *self, PyObject *args)
+{
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, tpc_commit);
+    EXC_IF_TPC_NOT_SUPPORTED(self);
+
+    return _psyco_conn_tpc_finish(self, args,
+                                  conn_commit, "COMMIT PREPARED");
+}
+
+#define psyco_conn_tpc_rollback_doc \
+"tpc_rollback([xid]) -- abort a transaction previously prepared."
+
+static PyObject *
+psyco_conn_tpc_rollback(connectionObject *self, PyObject *args)
+{
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, tpc_rollback);
+    EXC_IF_TPC_NOT_SUPPORTED(self);
+
+    return _psyco_conn_tpc_finish(self, args,
+                                  conn_rollback, "ROLLBACK PREPARED");
+}
+
+#define psyco_conn_tpc_recover_doc \
+"tpc_recover() -- returns a list of pending transaction IDs."
+
+static PyObject *
+psyco_conn_tpc_recover(connectionObject *self, PyObject *args)
+{
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, tpc_recover);
+    EXC_IF_TPC_PREPARED(self, tpc_recover);
+    EXC_IF_TPC_NOT_SUPPORTED(self);
+
+    return conn_tpc_recover(self);
+}
+
+
 #ifdef PSYCOPG_EXTENSIONS
 
 /* set_isolation_level method - switch connection isolation level */
@@ -163,22 +395,21 @@ psyco_conn_set_isolation_level(connectionObject *self, PyObject *args)
     int level = 1;
 
     EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, set_isolation_level);
+    EXC_IF_TPC_PREPARED(self, set_isolation_level);
 
     if (!PyArg_ParseTuple(args, "i", &level)) return NULL;
 
     if (level < 0 || level > 2) {
         PyErr_SetString(PyExc_ValueError,
-                        "isolation level out of bounds (0,3)");
+            "isolation level must be between 0 and 2");
         return NULL;
     }
 
     if (conn_switch_isolation_level(self, level) < 0) {
-        PyErr_SetString(OperationalError,
-                        PQerrorMessage(self->pgconn));
         return NULL;
     }
 
-
     Py_INCREF(Py_None);
     return Py_None;
 }
@@ -191,33 +422,20 @@ psyco_conn_set_isolation_level(connectionObject *self, PyObject *args)
 static PyObject *
 psyco_conn_set_client_encoding(connectionObject *self, PyObject *args)
 {
-    const char *enc = NULL;
-    char *buffer;
-    size_t i, j;
+    const char *enc;
+    PyObject *rv = NULL;
 
     EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, set_client_encoding);
+    EXC_IF_TPC_PREPARED(self, set_client_encoding);
 
     if (!PyArg_ParseTuple(args, "s", &enc)) return NULL;
 
-    /* convert to upper case and remove '-' and '_' from string */
-    buffer = PyMem_Malloc(strlen(enc)+1);
-    for (i=j=0 ; i < strlen(enc) ; i++) {
-        if (enc[i] == '_' || enc[i] == '-')
-            continue;
-        else
-            buffer[j++] = toupper(enc[i]);
-    }
-    buffer[j] = '\0';
-
-    if (conn_set_client_encoding(self, buffer) == 0) {
-        PyMem_Free(buffer);
+    if (conn_set_client_encoding(self, enc) == 0) {
         Py_INCREF(Py_None);
-        return Py_None;
-    }
-    else {
-        PyMem_Free(buffer);
-        return NULL;
+        rv = Py_None;
     }
+    return rv;
 }
 
 /* get_transaction_status method - Get backend transaction status */
@@ -230,8 +448,6 @@ psyco_conn_get_transaction_status(connectionObject *self, PyObject *args)
 {
     EXC_IF_CONN_CLOSED(self);
 
-    if (!PyArg_ParseTuple(args, "")) return NULL;
-
     return PyInt_FromLong((long)PQtransactionStatus(self->pgconn));
 }
 
@@ -261,14 +477,14 @@ psyco_conn_get_parameter_status(connectionObject *self, PyObject *args)
         Py_INCREF(Py_None);
         return Py_None;
     }
-    return PyString_FromString(val);
+    return conn_text_from_chars(self, val);
 }
 
 
 /* lobject method - allocate a new lobject */
 
 #define psyco_conn_lobject_doc \
-"cursor(oid=0, mode=0, new_oid=0, new_file=None,\n"                         \
+"lobject(oid=0, mode=0, new_oid=0, new_file=None,\n"                        \
 "       lobject_factory=extensions.lobject) -- new lobject\n\n"             \
 "Return a new lobject.\n\nThe ``lobject_factory`` argument can be used\n"   \
 "to create non-standard lobjects by passing a class different from the\n"   \
@@ -280,53 +496,37 @@ static PyObject *
 psyco_conn_lobject(connectionObject *self, PyObject *args, PyObject *keywds)
 {
     Oid oid=InvalidOid, new_oid=InvalidOid;
-    char *smode = NULL, *new_file = NULL;
-    int mode=0;
-    PyObject *obj, *factory = NULL;
+    char *new_file = NULL;
+    const char *smode = "";
+    PyObject *factory = (PyObject *)&lobjectType;
+    PyObject *obj;
 
     static char *kwlist[] = {"oid", "mode", "new_oid", "new_file",
                              "cursor_factory", NULL};
-    
+
     if (!PyArg_ParseTupleAndKeywords(args, keywds, "|izizO", kwlist,
-                                     &oid, &smode, &new_oid, &new_file, 
+                                     &oid, &smode, &new_oid, &new_file,
                                      &factory)) {
         return NULL;
     }
 
     EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, lobject);
+    EXC_IF_GREEN(lobject);
+    EXC_IF_TPC_PREPARED(self, lobject);
 
     Dprintf("psyco_conn_lobject: new lobject for connection at %p", self);
     Dprintf("psyco_conn_lobject:     parameters: oid = %d, mode = %s",
             oid, smode);
     Dprintf("psyco_conn_lobject:     parameters: new_oid = %d, new_file = %s",
             new_oid, new_file);
-    
-    /* build a mode number out of the mode string: right now we only accept
-       'r', 'w' and 'rw' (but note that 'w' implies 'rw' because PostgreSQL
-       backend does that. */
-    if (smode) {
-        if (strncmp("rw", smode, 2) == 0)
-            mode = INV_READ+INV_WRITE;
-        else if (smode[0] == 'r')
-           mode = INV_READ;
-        else if (smode[0] == 'w')
-           mode = INV_WRITE;
-        else if (smode[0] == 'n')
-           mode = -1;
-        else {
-            PyErr_SetString(PyExc_TypeError,
-                "mode should be one of 'r', 'w' or 'rw'");
-            return NULL;
-        }
-    }
 
-    if (factory == NULL) factory = (PyObject *)&lobjectType;
     if (new_file)
-        obj = PyObject_CallFunction(factory, "Oiiis", 
-            self, oid, mode, new_oid, new_file);
+        obj = PyObject_CallFunction(factory, "Oisis",
+            self, oid, smode, new_oid, new_file);
     else
-        obj = PyObject_CallFunction(factory, "Oiii",
-            self, oid, mode, new_oid);
+        obj = PyObject_CallFunction(factory, "Oisi",
+            self, oid, smode, new_oid);
 
     if (obj == NULL) return NULL;
     if (PyObject_IsInstance(obj, (PyObject *)&lobjectType) == 0) {
@@ -335,10 +535,10 @@ psyco_conn_lobject(connectionObject *self, PyObject *args, PyObject *keywds)
         Py_DECREF(obj);
         return NULL;
     }
-    
+
     Dprintf("psyco_conn_lobject: new lobject at %p: refcnt = "
             FORMAT_CODE_PY_SSIZE_T,
-            obj, obj->ob_refcnt);
+            obj, Py_REFCNT(obj));
     return obj;
 }
 
@@ -366,6 +566,7 @@ psyco_conn_reset(connectionObject *self)
     int res;
 
     EXC_IF_CONN_CLOSED(self);
+    EXC_IF_CONN_ASYNC(self, reset);
 
     if (pq_reset(self) < 0)
         return NULL;
@@ -378,8 +579,6 @@ psyco_conn_reset(connectionObject *self)
     return Py_None;
 }
 
-#endif
-
 static PyObject *
 psyco_conn_get_exception(PyObject *self, void *closure)
 {
@@ -389,6 +588,108 @@ psyco_conn_get_exception(PyObject *self, void *closure)
     return exception;
 }
 
+static PyObject *
+psyco_conn_poll(connectionObject *self)
+{
+    int res;
+
+    EXC_IF_CONN_CLOSED(self);
+
+    res = conn_poll(self);
+    if (res != PSYCO_POLL_ERROR || !PyErr_Occurred()) {
+        return PyInt_FromLong(res);
+    } else {
+        /* There is an error and an exception is already in place */
+        return NULL;
+    }
+}
+
+
+/* extension: fileno - return the file descriptor of the connection */
+
+#define psyco_conn_fileno_doc \
+"fileno() -> int -- Return file descriptor associated to database connection."
+
+static PyObject *
+psyco_conn_fileno(connectionObject *self)
+{
+    long int socket;
+
+    EXC_IF_CONN_CLOSED(self);
+
+    socket = (long int)PQsocket(self->pgconn);
+
+    return PyInt_FromLong(socket);
+}
+
+
+/* extension: isexecuting - check for asynchronous operations */
+
+#define psyco_conn_isexecuting_doc                           \
+"isexecuting() -> bool -- Return True if the connection is " \
+ "executing an asynchronous operation."
+
+static PyObject *
+psyco_conn_isexecuting(connectionObject *self)
+{
+    /* synchronous connections will always return False */
+    if (self->async == 0) {
+        Py_INCREF(Py_False);
+        return Py_False;
+    }
+
+    /* check if the connection is still being built */
+    if (self->status != CONN_STATUS_READY) {
+        Py_INCREF(Py_True);
+        return Py_True;
+    }
+
+    /* check if there is a query being executed */
+    if (self->async_cursor != NULL) {
+        Py_INCREF(Py_True);
+        return Py_True;
+    }
+
+    /* otherwise it's not executing */
+    Py_INCREF(Py_False);
+    return Py_False;
+}
+
+
+/* extension: cancel - cancel the current operation */
+
+#define psyco_conn_cancel_doc                           \
+"cancel() -- cancel the current operation"
+
+static PyObject *
+psyco_conn_cancel(connectionObject *self)
+{
+    char errbuf[256];
+
+    EXC_IF_CONN_CLOSED(self);
+    EXC_IF_TPC_PREPARED(self, cancel);
+
+    /* do not allow cancellation while the connection is being built */
+    Dprintf("psyco_conn_cancel: cancelling with key %p", self->cancel);
+    if (self->status != CONN_STATUS_READY &&
+        self->status != CONN_STATUS_BEGIN) {
+        PyErr_SetString(OperationalError,
+                        "asynchronous connection attempt underway");
+        return NULL;
+    }
+
+    if (PQcancel(self->cancel, errbuf, sizeof(errbuf)) == 0) {
+        Dprintf("psyco_conn_cancel: cancelling failed: %s", errbuf);
+        PyErr_SetString(OperationalError, errbuf);
+        return NULL;
+    }
+    Py_INCREF(Py_None);
+    return Py_None;
+}
+
+#endif  /* PSYCOPG_EXTENSIONS */
+
+
 /** the connection object **/
 
 
@@ -398,18 +699,30 @@ static struct PyMethodDef connectionObject_methods[] = {
     {"cursor", (PyCFunction)psyco_conn_cursor,
      METH_VARARGS|METH_KEYWORDS, psyco_conn_cursor_doc},
     {"close", (PyCFunction)psyco_conn_close,
-     METH_VARARGS, psyco_conn_close_doc},
+     METH_NOARGS, psyco_conn_close_doc},
     {"commit", (PyCFunction)psyco_conn_commit,
-     METH_VARARGS, psyco_conn_commit_doc},
+     METH_NOARGS, psyco_conn_commit_doc},
     {"rollback", (PyCFunction)psyco_conn_rollback,
-     METH_VARARGS, psyco_conn_rollback_doc},
+     METH_NOARGS, psyco_conn_rollback_doc},
+    {"xid", (PyCFunction)psyco_conn_xid,
+     METH_VARARGS|METH_KEYWORDS, psyco_conn_xid_doc},
+    {"tpc_begin", (PyCFunction)psyco_conn_tpc_begin,
+     METH_VARARGS, psyco_conn_tpc_begin_doc},
+    {"tpc_prepare", (PyCFunction)psyco_conn_tpc_prepare,
+     METH_NOARGS, psyco_conn_tpc_prepare_doc},
+    {"tpc_commit", (PyCFunction)psyco_conn_tpc_commit,
+     METH_VARARGS, psyco_conn_tpc_commit_doc},
+    {"tpc_rollback", (PyCFunction)psyco_conn_tpc_rollback,
+     METH_VARARGS, psyco_conn_tpc_rollback_doc},
+    {"tpc_recover", (PyCFunction)psyco_conn_tpc_recover,
+     METH_NOARGS, psyco_conn_tpc_recover_doc},
 #ifdef PSYCOPG_EXTENSIONS
     {"set_isolation_level", (PyCFunction)psyco_conn_set_isolation_level,
      METH_VARARGS, psyco_conn_set_isolation_level_doc},
     {"set_client_encoding", (PyCFunction)psyco_conn_set_client_encoding,
      METH_VARARGS, psyco_conn_set_client_encoding_doc},
     {"get_transaction_status", (PyCFunction)psyco_conn_get_transaction_status,
-     METH_VARARGS, psyco_conn_get_transaction_status_doc},
+     METH_NOARGS, psyco_conn_get_transaction_status_doc},
     {"get_parameter_status", (PyCFunction)psyco_conn_get_parameter_status,
      METH_VARARGS, psyco_conn_get_parameter_status_doc},
     {"get_backend_pid", (PyCFunction)psyco_conn_get_backend_pid,
@@ -418,6 +731,14 @@ static struct PyMethodDef connectionObject_methods[] = {
      METH_VARARGS|METH_KEYWORDS, psyco_conn_lobject_doc},
     {"reset", (PyCFunction)psyco_conn_reset,
      METH_NOARGS, psyco_conn_reset_doc},
+    {"poll", (PyCFunction)psyco_conn_poll,
+     METH_NOARGS, psyco_conn_lobject_doc},
+    {"fileno", (PyCFunction)psyco_conn_fileno,
+     METH_NOARGS, psyco_conn_fileno_doc},
+    {"isexecuting", (PyCFunction)psyco_conn_isexecuting,
+     METH_NOARGS, psyco_conn_isexecuting_doc},
+    {"cancel", (PyCFunction)psyco_conn_cancel,
+     METH_NOARGS, psyco_conn_cancel_doc},
 #endif
     {NULL}
 };
@@ -426,29 +747,31 @@ static struct PyMethodDef connectionObject_methods[] = {
 
 static struct PyMemberDef connectionObject_members[] = {
 #ifdef PSYCOPG_EXTENSIONS
-    {"closed", T_LONG, offsetof(connectionObject, closed), RO,
+    {"closed", T_LONG, offsetof(connectionObject, closed), READONLY,
         "True if the connection is closed."},
     {"isolation_level", T_LONG,
-        offsetof(connectionObject, isolation_level), RO,
+        offsetof(connectionObject, isolation_level), READONLY,
         "The current isolation level."},
-    {"encoding", T_STRING, offsetof(connectionObject, encoding), RO,
+    {"encoding", T_STRING, offsetof(connectionObject, encoding), READONLY,
         "The current client encoding."},
-    {"notices", T_OBJECT, offsetof(connectionObject, notice_list), RO},
-    {"notifies", T_OBJECT, offsetof(connectionObject, notifies), RO},
-    {"dsn", T_STRING, offsetof(connectionObject, dsn), RO,
+    {"notices", T_OBJECT, offsetof(connectionObject, notice_list), READONLY},
+    {"notifies", T_OBJECT, offsetof(connectionObject, notifies), READONLY},
+    {"dsn", T_STRING, offsetof(connectionObject, dsn), READONLY,
         "The current connection string."},
+    {"async", T_LONG, offsetof(connectionObject, async), READONLY,
+        "True if the connection is asynchronous."},
     {"status", T_INT,
-        offsetof(connectionObject, status), RO,
+        offsetof(connectionObject, status), READONLY,
         "The current transaction status."},
-    {"string_types", T_OBJECT, offsetof(connectionObject, string_types), RO,
+    {"string_types", T_OBJECT, offsetof(connectionObject, string_types), READONLY,
         "A set of typecasters to convert textual values."},
-    {"binary_types", T_OBJECT, offsetof(connectionObject, binary_types), RO,
+    {"binary_types", T_OBJECT, offsetof(connectionObject, binary_types), READONLY,
         "A set of typecasters to convert binary values."},
     {"protocol_version", T_INT,
-        offsetof(connectionObject, protocol), RO,
-        "Protocol version (2 or 3) used for this connection."},
+        offsetof(connectionObject, protocol), READONLY,
+        "Protocol version used for this connection. Currently always 3."},
     {"server_version", T_INT,
-        offsetof(connectionObject, server_version), RO,
+        offsetof(connectionObject, server_version), READONLY,
         "Server version."},
 #endif
     {NULL}
@@ -476,40 +799,39 @@ static struct PyGetSetDef connectionObject_getsets[] = {
 /* initialization and finalization methods */
 
 static int
-connection_setup(connectionObject *self, const char *dsn)
+connection_setup(connectionObject *self, const char *dsn, long int async)
 {
     char *pos;
-    int res;
+    int res = -1;
 
-    Dprintf("connection_setup: init connection object at %p, refcnt = "
-        FORMAT_CODE_PY_SSIZE_T,
-        self, ((PyObject *)self)->ob_refcnt
+    Dprintf("connection_setup: init connection object at %p, "
+	    "async %ld, refcnt = " FORMAT_CODE_PY_SSIZE_T,
+            self, async, Py_REFCNT(self)
       );
 
-    self->dsn = strdup(dsn);
-    self->notice_list = PyList_New(0);
-    self->notifies = PyList_New(0);
-    self->closed = 0;
-    self->status = CONN_STATUS_READY;
-    self->critical = NULL;
-    self->async_cursor = NULL;
-    self->pgconn = NULL;
-    self->mark = 0;
-    self->string_types = PyDict_New();
-    self->binary_types = PyDict_New();
-    self->notice_pending = NULL;
-    self->encoding = NULL;
+    if (!(self->dsn = strdup(dsn))) {
+        PyErr_NoMemory();
+        goto exit;
+    }
+    if (!(self->notice_list = PyList_New(0))) { goto exit; }
+    if (!(self->notifies = PyList_New(0))) { goto exit; }
+    self->async = async;
+    self->status = CONN_STATUS_SETUP;
+    self->async_status = ASYNC_DONE;
+    if (!(self->string_types = PyDict_New())) { goto exit; }
+    if (!(self->binary_types = PyDict_New())) { goto exit; }
+    /* other fields have been zeroed by tp_alloc */
 
     pthread_mutex_init(&(self->lock), NULL);
 
-    if (conn_connect(self) != 0) {
+    if (conn_connect(self, async) != 0) {
         Dprintf("connection_init: FAILED");
-        res = -1;
+        goto exit;
     }
     else {
         Dprintf("connection_setup: good connection object at %p, refcnt = "
             FORMAT_CODE_PY_SSIZE_T,
-            self, ((PyObject *)self)->ob_refcnt
+            self, Py_REFCNT(self)
           );
         res = 0;
     }
@@ -521,6 +843,7 @@ connection_setup(connectionObject *self, const char *dsn)
             *pos = 'x';
     }
 
+exit:
     return res;
 }
 
@@ -528,17 +851,23 @@ static void
 connection_dealloc(PyObject* obj)
 {
     connectionObject *self = (connectionObject *)obj;
-    
+
+    if (self->weakreflist) {
+        PyObject_ClearWeakRefs(obj);
+    }
+
     PyObject_GC_UnTrack(self);
 
     if (self->closed == 0) conn_close(self);
-    
+
     conn_notice_clean(self);
 
     if (self->dsn) free(self->dsn);
-    if (self->encoding) free(self->encoding);
+    PyMem_Free(self->encoding);
+    PyMem_Free(self->codec);
     if (self->critical) free(self->critical);
 
+    Py_CLEAR(self->tpc_xid);
     Py_CLEAR(self->async_cursor);
     Py_CLEAR(self->notice_list);
     Py_CLEAR(self->notice_filter);
@@ -550,21 +879,23 @@ connection_dealloc(PyObject* obj)
 
     Dprintf("connection_dealloc: deleted connection object at %p, refcnt = "
         FORMAT_CODE_PY_SSIZE_T,
-        obj, obj->ob_refcnt
+        obj, Py_REFCNT(obj)
       );
 
-    obj->ob_type->tp_free(obj);
+    Py_TYPE(obj)->tp_free(obj);
 }
 
 static int
 connection_init(PyObject *obj, PyObject *args, PyObject *kwds)
 {
     const char *dsn;
+    long int async = 0;
+    static char *kwlist[] = {"dsn", "async", NULL};
 
-    if (!PyArg_ParseTuple(args, "s", &dsn))
+    if (!PyArg_ParseTupleAndKeywords(args, kwds, "s|l", kwlist, &dsn, &async))
         return -1;
 
-    return connection_setup((connectionObject *)obj, dsn);
+    return connection_setup((connectionObject *)obj, dsn, async);
 }
 
 static PyObject *
@@ -590,6 +921,7 @@ connection_repr(connectionObject *self)
 static int
 connection_traverse(connectionObject *self, visitproc visit, void *arg)
 {
+    Py_VISIT((PyObject *)(self->tpc_xid));
     Py_VISIT(self->async_cursor);
     Py_VISIT(self->notice_list);
     Py_VISIT(self->notice_filter);
@@ -610,8 +942,7 @@ connection_traverse(connectionObject *self, visitproc visit, void *arg)
 "    ProgrammingError, IntegrityError, DataError, NotSupportedError"
 
 PyTypeObject connectionType = {
-    PyObject_HEAD_INIT(NULL)
-    0,
+    PyVarObject_HEAD_INIT(NULL, 0)
     "psycopg2._psycopg.connection",
     sizeof(connectionObject),
     0,
@@ -632,14 +963,16 @@ PyTypeObject connectionType = {
     0,          /*tp_setattro*/
     0,          /*tp_as_buffer*/
 
-    Py_TPFLAGS_DEFAULT|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
+    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_BASETYPE | Py_TPFLAGS_HAVE_GC |
+        Py_TPFLAGS_HAVE_WEAKREFS,
+                /*tp_flags*/
     connectionType_doc, /*tp_doc*/
 
     (traverseproc)connection_traverse, /*tp_traverse*/
     0,          /*tp_clear*/
 
     0,          /*tp_richcompare*/
-    0,          /*tp_weaklistoffset*/
+    offsetof(connectionObject, weakreflist), /* tp_weaklistoffset */
 
     0,          /*tp_iter*/
     0,          /*tp_iternext*/
